use crate::vm::dm_type::DmType;
use crate::vm::dvm_value::DvmValue;
use crate::vm::object::DvmObject;
use std::rc::Rc;

unsafe fn read_field(data_pointer: *const u8, dm_type: &DmType) -> DvmValue {
    match dm_type {
        DmType::Byte => DvmValue::Byte(data_pointer.read()),
        DmType::Int => DvmValue::Int(data_pointer.cast::<i32>().read()),
        DmType::Long => DvmValue::Long(data_pointer.cast::<i64>().read()),
        DmType::USize => DvmValue::USize(data_pointer.cast::<usize>().read()),
        DmType::Double => DvmValue::Double(data_pointer.cast::<f64>().read()),
        DmType::Boolean => DvmValue::Boolean(data_pointer.cast::<bool>().read()),
        DmType::Object => DvmValue::Object(data_pointer.cast::<Rc<DvmObject>>().read()),
    }
}

pub fn read_field_by_name(field_name: &str, self_object: &DvmObject) -> DvmValue {
    let (field_index, field) = self_object
        .implementation()
        .fields
        .iter()
        .enumerate()
        .find(|(_index, field)| field.name() == field_name)
        .expect(&format!(
            "Cannot find field {} in {}",
            field_name,
            self_object.implementation().fqn
        ));
    let data_pointer = self_object.data()[field_index];
    unsafe { read_field(data_pointer, field.dm_type()) }
}

pub fn read_field_by_index(index: usize, self_object: &DvmObject) -> DvmValue {
    let field = &self_object.implementation().fields[index];
    let data_pointer = self_object.data()[index];
    unsafe { read_field(data_pointer, field.dm_type()) }
}

unsafe fn write_field(data_pointer: *mut u8, value: DvmValue) {
    match value {
        DvmValue::Byte(b) => {
            data_pointer.write(b);
        }
        DvmValue::Int(i) => {
            data_pointer.cast::<i32>().write(i);
        }
        DvmValue::Long(l) => {
            data_pointer.cast::<i64>().write(l);
        }
        DvmValue::USize(u) => {
            data_pointer.cast::<usize>().write(u);
        }
        DvmValue::Double(d) => {
            data_pointer.cast::<f64>().write(d);
        }
        DvmValue::Boolean(b) => {
            data_pointer.cast::<bool>().write(b);
        }
        DvmValue::Object(o) => {
            data_pointer.cast::<Rc<DvmObject>>().write(o);
        }
        DvmValue::Uninit => panic!("Cannot write DvmValue::Uninit to field."),
        DvmValue::Void => panic!("Cannot write DvmValue::Void to field."),
    }
}

pub fn write_field_by_name(field_name: &str, self_object: &DvmObject, value: DvmValue) {
    let field_index = self_object
        .implementation()
        .fields
        .iter()
        .position(|field| field.name() == field_name)
        .expect(&format!(
            "Cannot find field {} in {}",
            field_name,
            self_object.implementation().fqn
        ));
    let data_pointer = self_object.data()[field_index];
    unsafe { write_field(data_pointer, value) };
}

pub fn write_field_by_index(index: usize, self_object: &DvmObject, value: DvmValue) {
    let data_pointer = self_object.data()[index];
    unsafe { write_field(data_pointer, value) };
}
